import pytest
from slate import Span

from zenith.markup import (
    zml,
    zml_get_spans,
    zml_alias,
    zml_macro,
    zml_pre_process,
    GLOBAL_CONTEXT,
)


def test_markup_str_parse():
    @zml_macro
    def upper(text: str) -> str:
        return text.upper()

    assert (
        zml("[bold underline 61]Hello [141 /bold]There")
        == "\x1b[38;5;61;1;4mHello \x1b[22m\x1b[38;5;141mThere\x1b[0m"
    )

    assert zml("[38]Hello [!upper]There") == "\x1b[38;5;38mHello THERE\x1b[0m"

    assert zml(
        "Click [141 underline ~https://google.com]me[61] and me[/~ / bold red] NOW"
    ) == (
        "Click \x1b]8;;https://google.com\x1b\\\x1b[38;5;141;4mme"
        + "\x1b]8;;https://google.com\x1b\\\x1b[38;5;61m and me\x1b[0m"
        + "\x1b]8;;\x1b\\\x1b[38;2;255;0;0;1m NOW\x1b[0m"
    )

    assert zml("The never-ending [~https://google.com]URI") == (
        "The never-ending \x1b]8;;https://google.com\x1b\\URI\x1b]8;;\x1b\\"
    ), repr(zml("The never-ending [~https://google.com]URI"))

    assert zml("[blue][@red]This is hard to parse") == (
        "\x1b[38;2;0;0;255;48;2;255;0;0mThis is hard to parse\x1b[0m"
    )


def test_markup_parse():
    assert list(zml_get_spans("[bold italic]Hello")) == [
        Span("Hello", bold=True, italic=True),
    ]

    assert list(zml_get_spans("[bold italic]Hello[/bold]Not bold")) == [
        Span("Hello", bold=True, italic=True),
        Span("Not bold", italic=True),
    ]


def test_markup_color():
    # 0-7 indexed colors
    assert list(zml_get_spans("[@0 1]Test[@1 0]Other")) == [
        Span("Test", background="40", foreground="31"),
        Span("Other", background="41", foreground="30"),
    ]

    # 8-15 indexed colors
    assert list(zml_get_spans("[@9 15]Test[@14 10]Other")) == [
        Span("Test", background="101", foreground="97"),
        Span("Other", background="106", foreground="92"),
    ]

    # 16-256 indexed colors
    assert list(zml_get_spans("[@141 61]Test[@54 78]Other")) == [
        Span("Test", background="48;5;141", foreground="38;5;61"),
        Span("Other", background="48;5;54", foreground="38;5;78"),
    ]

    # OOB indexed color
    with pytest.raises(ValueError):
        list(zml_get_spans("[@333 231]Please don't parse..."))

    # CSS colors
    assert list(zml_get_spans("[@lavender cadetblue]Pretty colors")) == [
        Span(
            "Pretty colors", foreground="38;2;95;158;160", background="48;2;230;230;250"
        )
    ]

    # HEX colors
    assert list(zml_get_spans("[@#212121 #dedede]Nice contrast")) == [
        Span("Nice contrast", foreground="38;2;222;222;222", background="48;2;33;33;33")
    ]

    # RGB colors
    assert list(zml_get_spans("[@11;22;123 123;22;11]What even are these colors")) == [
        Span(
            "What even are these colors",
            foreground="38;2;123;22;11",
            background="48;2;11;22;123",
        )
    ]


def test_markup_auto_foreground():
    assert list(zml_get_spans("[@#FFFFFF]test")) == [
        Span("test", background="48;2;255;255;255", foreground="38;2;35;35;35")
    ]

    assert list(zml_get_spans("[@0]White[@7]Black[@160]White[@116]Black")) == [
        Span("White", background="40", foreground="38;2;245;245;245"),
        Span("Black", background="47", foreground="38;2;35;35;35"),
        Span("White", background="48;5;160", foreground="38;2;245;245;245"),
        Span("Black", background="48;5;116", foreground="38;2;35;35;35"),
    ]


def test_markup_macros():
    @zml_macro
    def upper(text: str) -> str:
        return text.upper()

    @zml_macro
    def conceal(text: str) -> str:
        return "*" * (len(text) - 1) + text[-1]

    assert (
        zml_pre_process("[!upper 141]Test[/!upper /fg !conceal bold]other test")
        == "[141]TEST[/fg bold]*********t"
    )

    with pytest.raises(ValueError):
        zml_pre_process("[!undefined]Test")

    with pytest.raises(ValueError):
        zml_pre_process("[undefined]Test")

    with pytest.raises(ValueError):
        zml_pre_process("[/!upper]Test")


def test_markup_aliases():
    zml_alias(a="b", test="141 bold")

    assert zml_pre_process("[test italic]What is this?") == (
        "[141 bold italic]What is this?"
    )

    @zml_macro
    def upper(text: str) -> str:
        return text.upper()

    zml_alias(complex_with_macro="!upper lavender")

    assert zml_pre_process("[complex-with-macro]test") == "[lavender]TEST"

    zml_alias(complex_with_hyperlink="~https://google.com underline slategray")

    assert (
        zml_pre_process("[complex-with-hyperlink]test")
        == "[~https://google.com underline slategray]test"
    )


def test_markup_hyperlink():
    assert list(
        zml_get_spans("[~https://google.com]Test [bold]me [/~]no more link")
    ) == [
        Span("Test ", hyperlink="https://google.com"),
        Span("me ", bold=True, hyperlink="https://google.com"),
        Span("no more link", bold=True),
    ]


def test_markup_parse_plain():
    assert list(zml_get_spans("Test")) == [Span("Test")]



from slate.color import Color
from zenith.markup import zml_context
from zenith.palette import Palette, analogous, tetradic, triadic


def test_palette_triadic():
    pal = Palette.from_hex("#FABCDE", strategy=triadic)

    assert pal.primary == Color.from_hex("#FABCDE")
    assert pal.secondary == pal.primary.hue_shift(1 / 3)
    assert pal.tertiary == pal.primary.hue_shift(2 / 3)
    assert pal.quaternary == pal.primary.complement


def test_palette_analogous():
    pal = Palette.from_hex("#EDCBAF", strategy=analogous)

    assert pal.primary == Color.from_hex("#EDCBAF")
    assert pal.secondary == pal.primary.hue_shift(1 / 12)
    assert pal.tertiary == pal.primary.hue_shift(2 / 12)
    assert pal.quaternary == pal.primary.complement


def test_palette_tetradic():
    pal = Palette.from_hex("#F45A11", strategy=tetradic)

    assert pal.primary == Color.from_hex("#F45A11")
    assert pal.secondary == pal.primary.hue_shift(1 / 4)
    assert pal.tertiary == pal.primary.hue_shift(2 / 4)
    assert pal.quaternary == pal.primary.hue_shift(3 / 4)


def test_palette_alias():
    pal = Palette.from_hex("#42DFBC")
    ctx = zml_context()

    pal.alias(ctx=ctx)

    assert ctx["aliases"]["error-2"] == pal.error.darken(2).hex
    assert ctx["aliases"]["surface3"] == pal.surface3.hex
    assert ctx["aliases"]["success+3"] == pal.success.lighten(3).hex

    assert "success+4" not in ctx["aliases"]
    assert "primary-5" not in ctx["aliases"]



"""A simple implementation for a Least-Recently-Used cache object.

We need this because Python's `lru_cache` doesn't support non-hashable objects,
and adding support (through another decorator) is messy and impossible to read.
"""

from __future__ import annotations

from collections import OrderedDict
from dataclasses import dataclass, field
from typing import Generic, Hashable, TypeVar

__all__ = ["LRUCache"]

CacheItemType = TypeVar("CacheItem")


@dataclass
class LRUCache(Generic[CacheItemType]):
    """A Least-Recently-Used cache.

    Usage:

    ```python3
    cache = LRUCache(1024)

    def cached_function(a: int, b: int) -> float:
        key = please_hash(a, b)

        if key in cache:
            return cache[key]

        # Expensive calculation
        result = ...

        cache[key] = result

        return result
    ```
    """

    _capacity: int
    _cache: OrderedDict[Hashable, CacheItemType] = field(
        init=False, default_factory=OrderedDict
    )

    def __contains__(self, key: Hashable) -> bool:
        return key in self._cache

    def __getitem__(self, key: Hashable) -> CacheItemType:
        if key not in self._cache:
            raise KeyError(f"Key {key!r} not found in cache.")

        self._cache.move_to_end(key)
        return self._cache[key]

    def __setitem__(self, key: Hashable, value: CacheItemType) -> None:
        self._cache[key] = value
        self._cache.move_to_end(key)

        if len(self._cache) > self._capacity:
            self._cache.popitem(last=False)

    def get(
        self, key: Hashable, default: CacheItemType | None = None
    ) -> CacheItemType | None:
        return self._cache.get(key, default)



import pytest

from zenith.lru_cache import LRUCache


def test_lru_cache():
    cache = LRUCache(2)

    cache[0] = "something"

    assert 0 in cache
    assert cache[0] == "something"

    cache[1] = "something else"

    assert 1 in cache
    assert cache[1] == "something else"

    cache[2] = "third item"
    assert 1 in cache and 2 in cache and not 0 in cache

    with pytest.raises(KeyError):
        cache[4]



"""The Palette class and its built-in strategies."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Callable

from slate.color import Color

from .markup import MarkupContext, zml_alias

__all__ = [
    "triadic",
    "analogous",
    "tetradic",
    "PalettingFunction",
    "Palette",
]

PalettingFunction = Callable[[Color], tuple[Color, Color, Color, Color]]


def triadic(base: Color) -> tuple[Color, Color, Color, Color]:
    r"""Returns a complementary triangle including base and the base's complement.

    On a colorwheel:

           . 0 .
         .  /|\  .
        .  / | \  .
        . /  |  \ .
         2---+---1
           . 3 .

    """

    return (*base.triadic_group, base.complement)


def analogous(base: Color) -> tuple[Color, Color, Color, Color]:
    """Returns three colors next to eachother, including base.

    On a colorwheel:

           . 0 1
         .       2
        .         .
         .       .
           . 3 .

    """

    return (*base.analogous_group, base.complement)


def tetradic(base: Color) -> tuple[Color, Color, Color, Color]:
    """Returns a complementary square including base.

    On a colorwheel:

           . 0 .
         .       .
        3         1
         .       .
           . 2 .

    """

    return base.tetradic_group


DEFAULT_SURFACE = Color.black().blend_complement(0.2)
SURFACE_BLEND_ALPHA = 0.2

DEFAULT_SUCCESS = Color.from_hex("67eb7f")
DEFAULT_ERROR = Color.from_hex("eb7067")
DEFAULT_WARNING = Color.from_hex("ebe267")
SEMANTIC_BLEND_ALPHA = 0.3


@dataclass
class Palette:  # pylint: disable=too-many-instance-attributes
    """A palette containing 4 'main' colors and surface colors and 3 semantics ones.

    The primary color is the main color of an interface. Secondary, tertiary and
    quaternary colors are calculated based on the given strategy.

    Success, warning & error are colors representing the given semantic meanings; they
    are computed by tinting some defaults with the primary color.

    Additionally, there are surface colors defined for all 4 of the main ones; these are
    created by tinting a default surface color with the primary, secondary, tertiary and
    quaternary color respectively.
    """

    primary: Color
    secondary: Color = field(init=False)
    tertiary: Color = field(init=False)
    quaternary: Color = field(init=False)

    success: Color = field(init=False)
    warning: Color = field(init=False)
    error: Color = field(init=False)

    surface1: Color = field(init=False)
    surface2: Color = field(init=False)
    surface3: Color = field(init=False)
    surface4: Color = field(init=False)

    surface_blend_base: Color = DEFAULT_SURFACE
    success_blend_base: Color = DEFAULT_SUCCESS
    warning_blend_base: Color = DEFAULT_WARNING
    error_blend_base: Color = DEFAULT_ERROR

    strategy: PalettingFunction = triadic

    def __post_init__(self) -> None:
        primary = self.primary

        _, self.secondary, self.tertiary, self.quaternary = self.strategy(primary)

        self.success = self.success_blend_base.blend(primary, SEMANTIC_BLEND_ALPHA)
        self.warning = self.warning_blend_base.blend(primary, SEMANTIC_BLEND_ALPHA)
        self.error = self.error_blend_base.blend(primary, SEMANTIC_BLEND_ALPHA)

        self.surface1 = self.surface_blend_base.blend(
            self.primary,
            alpha=SURFACE_BLEND_ALPHA,
        )
        self.surface2 = self.surface_blend_base.blend(
            self.secondary,
            alpha=SURFACE_BLEND_ALPHA,
        )
        self.surface3 = self.surface_blend_base.blend(
            self.tertiary,
            alpha=SURFACE_BLEND_ALPHA,
        )
        self.surface4 = self.surface_blend_base.blend(
            self.quaternary,
            alpha=SURFACE_BLEND_ALPHA,
        )

        self.color_mapping = {
            "primary": self.primary,
            "secondary": self.secondary,
            "tertiary": self.tertiary,
            "quaternary": self.quaternary,
            "success": self.success,
            "warning": self.warning,
            "error": self.error,
            "surface1": self.surface1,
            "surface2": self.surface2,
            "surface3": self.surface3,
            "surface4": self.surface4,
        }

    @classmethod
    def from_hex(cls, primary: str, strategy: PalettingFunction = triadic) -> Palette:
        """Generates a palette from a CSS-style HEX color string."""

        return Palette(Color.from_hex(primary), strategy=strategy)

    @property
    def color_mapping(self) -> dict[str, Color]:
        """Returns a mapping of color names to their values.

        Primarily used for aliasing.
        """

        return self._color_mapping

    @color_mapping.setter
    def color_mapping(self, new: dict[str, Color]) -> None:
        """Sets the color mapping."""

        self._color_mapping = new

    def alias(
        self,
        ctx: ContextMapping | None = None,
        mapping: dict[str, Color] | None = None,
        shade_count: int = 3,
        shade_step: float = 0.1,
    ) -> None:
        """Sets up aliases for this palette's colors.

        Args:
            ctx: The markup context to alias into. Defaults to the global context.
            mapping: The name->color mapping used to set up the aliases. Defaults to
                `self.color_mapping`.
            shade_count: The number of shades that should be included on both the
                positive and negative side. These are aliased as `{color}{+/-step}`,
                like `primary+2` or `surface1-2`.
            shade_step: The step_size passed to `Color.lighten` or `Color.darken`,
                controls the distance between shades.
        """

        mapping = mapping or self.color_mapping

        for name, color in mapping.items():
            for i in range(-shade_count, shade_count + 1):
                if i == 0:
                    key = name
                    colorhex = color.hex

                elif i < 0:
                    key = f"{name}{i}"
                    colorhex = color.darken(-i, step_size=shade_step).hex

                else:
                    key = f"{name}+{i}"
                    colorhex = color.lighten(i, step_size=shade_step).hex

                zml_alias(**{key: colorhex, f"@{key}": f"@{colorhex}"}, ctx=ctx)

    def render(self) -> str:
        """Returns markup that shows off the palette.

        Note that this is done according to the current `color_mapping`.
        """

        length = len(max(self.color_mapping.keys(), key=len)) + 4

        lines = [[] for _ in range(7)]

        for name in self.color_mapping:
            for i, shade in enumerate([str(i) for i in range(-3, 4)]):
                if shade == "0":
                    display = name
                    identifier = name
                    shade = ""

                else:
                    display = ""
                    identifier = f"{name}{('+' if int(shade) > 0 else '') + shade}"

                lines[i].append(f"[@{identifier}]{display:^{length}}")

        return "\n".join(" ".join(line) for line in lines)



from __future__ import annotations

from dataclasses import dataclass

__all__ = [
    "ZmlError",
    "ZmlNameError",
    "ZmlSemanticsError",
]


def _generate_error_text(start: str, expected_type: str, tag: str, context: str) -> str:
    """Generates a generic error message."""

    buff = f"{start} {expected_type} {tag!r}"

    if context is None:
        return buff + "."

    return buff + f": {context}"


class ZmlError(BaseException):
    """An error raised from ZML."""


@dataclass
class ZmlNameError(ZmlError):
    """Raised when an unknown tag is used."""

    tag: str
    context: str | None = None
    expected_type: str = "tag"

    def __str__(self) -> str:
        return _generate_error_text(
            "Unknown", self.expected_type, self.tag, self.context
        )


@dataclass
class ZmlSemanticsError(ZmlError):
    """Raised when a semantic error occurs, such as unsetting a macro that isn't set."""

    tag: str
    context: str | None = None
    expected_type: str = "tag"

    def __str__(self) -> str:
        return _generate_error_text(
            "Invalid", self.expected_type, self.tag, self.context
        )



import time
import pytest

from zenith import zml, zml_alias, zml_macro, ZmlNameError, ZmlSemanticsError
from zenith.markup import parse_color


def test_markup_builtin_only():
    assert (
        result := zml("[bold 141]Hello [@61]There ")
    ) == "\x1b[38;5;141;1mHello \x1b[48;5;61mThere ", repr(result)

    assert (
        result := zml("[bold 141]Hello [/bold @61]There ")
    ) == "\x1b[38;5;141;1mHello \x1b[22m\x1b[48;5;61mThere ", repr(result)

    assert (
        result := zml("[bold 141]Hello [/]There ")
    ) == "\x1b[38;5;141;1mHello \x1b[0mThere ", repr(result)

    assert (
        result := zml("[~https://github.com]Hello There ")
    ) == "\x1b]8;;https://github.com\x1b\\Hello There \x1b]8;;\x1b\\", repr(result)

    assert (result := zml("[~https://github.com]Hello [red]There [/~]no link")) == (
        "\x1b]8;;https://github.com\x1b\\Hello "
        + "\x1b]8;;\x1b\\\x1b[38;2;255;0;0mThere no link"
    ), repr(result)

    assert (
        result := zml("[127]Hello [/fg]There ")
    ) == "\x1b[38;5;127mHello \x1b[39mThere ", repr(result)

    assert (
        result := zml("[127]Hello [/bg]There ")
    ) == "\x1b[38;5;127mHello There ", repr(result)


def test_markup_aliases():
    zml_alias(test="brown bold")

    assert (
        result := zml("[test @red]Hello")
    ) == "\x1b[38;2;165;42;42;48;2;255;0;0;1mHello", repr(result)

    assert (
        result := zml("[test]Hello[/test]Reset")
    ) == "\x1b[38;2;165;42;42;1mHello\x1b[22;39mReset", repr(result)


def test_markup_macros():
    @zml_macro
    def epoch(template: str, fmt: str = "%c") -> str:
        return template.format(time=f"the current time with {fmt=}")

    assert (
        result := zml("[!epoch]Hello at {time}[/!epoch]{time}")
    ) == "Hello at the current time with fmt='%c'{time}", repr(result)

    assert (
        result := zml("[!epoch(%M)]Hello at {time}[/!epoch]{time}")
    ) == "Hello at the current time with fmt='%M'{time}", repr(result)

    with pytest.raises(TypeError):
        zml("[!epoch(%m,%M)]test {time}")

    @zml_macro
    def debug(template: str, *args: str) -> str:
        return template.format(args=str(args))

    assert (result := zml("[!debug(arg1,arg2)]{args}")) == "('arg1', 'arg2')", repr(
        result
    )

    assert (
        result := zml("[!debug(arg1,arg2) red bold]{args}[/fg]other: {args}")
    ) == "\x1b[38;2;255;0;0;1m('arg1', 'arg2')\x1b[39mother: ('arg1', 'arg2')", repr(
        result
    )


def test_markup_auto_foreground():
    assert (
        result := zml("[@red]Test")
    ) == "\x1b[38;2;35;35;35;48;2;255;0;0mTest", repr(result)

    assert (
        result := zml("[@black]Test")
    ) == "\x1b[38;2;245;245;245;48;2;0;0;0mTest", repr(result)

    assert (
        result := zml("[@black black]Test")
    ) == "\x1b[38;2;0;0;0;48;2;0;0;0mTest", repr(result)

    assert (
        result := zml("[@yellow]Test")
    ) == "\x1b[38;2;35;35;35;48;2;255;255;0mTest", repr(result)

    assert (
        result := zml("[invert yellow]Test")
    ) == "\x1b[38;2;255;255;0;48;2;35;35;35;7mTest", repr(result)

    assert (
        (result := zml("[@yellow]Black[@black]White"))
        == "\x1b[38;2;35;35;35;48;2;255;255;0mBlack\x1b[38;2;245;245;245;48;2;0;0;0mWhite"
    ), repr(result)


def test_markup_colors():
    assert parse_color("red", False) == "38;2;255;0;0"
    assert parse_color("blue", True) == "48;2;0;0;255"
    assert parse_color("2", False) == "32"
    assert parse_color("9", False) == "91"
    assert parse_color("14", True) == "106"
    assert parse_color("78", True) == "48;5;78"
    assert (
        parse_color("#baebae", True)
        == parse_color("186;235;174", True)
        == "48;2;186;235;174"
    )


def test_markup_parse_errors():
    with pytest.raises(ZmlNameError):
        parse_color("256", False)

    with pytest.raises(ZmlNameError):
        zml("[not-a-tag]Hello")

    with pytest.raises(ZmlNameError):
        zml("[!not-a-macro]Hello")

    with pytest.raises(ZmlSemanticsError):
        zml("Test[/!not-a-macro]")

    @zml_macro
    def not_a_macro(text: str) -> str:
        return "Why did you call this"

    with pytest.raises(ZmlSemanticsError):
        zml("Test[/!not-a-macro]")



from __future__ import annotations

import re

from functools import lru_cache
from typing import Callable, TypedDict

from slate.span import UNSETTERS, SETTERS_TO_STYLES, Span
from slate.color import Color
from slate.color_info import NAMED_COLORS

from .exceptions import ZmlNameError, ZmlSemanticsError
from .lru_cache import LRUCache

__all__ = [
    "zml",
    "zml_alias",
    "zml_escape",
    "zml_macro",
    "combine_spans",
    "zml_get_spans",
    "zml_pre_process",
    "zml_context",
    "MarkupContext",
    "GLOBAL_CONTEXT",
]

MacroType = Callable[[str], str]

FULL_RESET = Span("FULL_RESET")
"""A sentinel value to be used for signifying a full style reset (CSI 0)."""

RE_MARKUP = re.compile(r"(?:\[([^\[\]]+)\])?([^\]\[]+)?")
RE_COLOR = re.compile(
    r"(?:^@?([\d]{1,3})$)|(?:@?#?([0-9a-fA-F]{6}))|(@?\d{1,3};\d{1,3};\d{1,3})"
)


class MarkupContext(TypedDict):
    """A ctx mapping that stores alias & macro information."""

    aliases: dict[str, str]
    macros: dict[str, MacroType]


def zml_context() -> MarkupContext:
    """Generates an empty MarkupContext object."""

    return {"aliases": {}, "macros": {}}


GLOBAL_CONTEXT = zml_context()


def zml_macro_setter(*, ctx: MarkupContext) -> Callable[[MacroType], MacroType]:
    """Returns a decorator that will define a macro in the given context."""

    def _define(macro: MacroType) -> MacroType:
        name = macro.__name__.replace("_", "-")

        ctx["macros"][f"!{name}"] = macro

    return _define


zml_macro = zml_macro_setter(ctx=GLOBAL_CONTEXT)
"""A macro setter for the global context."""


def _find_unsetter(tag: str) -> str:
    """Finds the unsetter for the given tag."""

    if tag.startswith("!"):
        return tag

    key = tag

    if RE_COLOR.match(tag) or tag in NAMED_COLORS:
        key = "bg" if tag.startswith("@") else "fg"

    return f"/{key}"


def zml_alias(
    *, ctx: MarkupContext | None = None, assign_unsetter: bool = True, **pairs: str
) -> None:
    """Aliases each item of the given pairs."""

    if assign_unsetter:
        for key, value in pairs.copy().items():
            pairs[f"/{key}"] = " ".join(_find_unsetter(part) for part in value.split())

    ctx = ctx or GLOBAL_CONTEXT

    aliases = ctx["aliases"]

    for key, value in pairs.items():
        aliases[key.replace("_", "-")] = value


class StyleMap(TypedDict):
    """A type to keep track of span styles."""

    bold: bool
    dim: bool
    italic: bool
    underline: bool
    blink: bool
    fast_blink: bool
    invert: bool
    conceal: bool
    strike: bool

    foreground: Color | None
    background: Color | None
    hyperlink: str


def _get_style_map() -> StyleMap:
    """Generates an empty StyleMap object."""

    return {
        "bold": False,
        "dim": False,
        "italic": False,
        "underline": False,
        "blink": False,
        "fast_blink": False,
        "invert": False,
        "conceal": False,
        "strike": False,
        "foreground": None,
        "background": None,
        "hyperlink": "",
    }


def _apply_auto_foreground(styles: StyleMap) -> bool:
    """Determines whether automatic foreground can be applied to the style map."""

    foreground = styles["foreground"]
    background = styles["background"]

    invert = styles["invert"]
    if invert:
        foreground, background = background, foreground

    if foreground is None and background is not None:
        new = background.contrast.as_background(invert)

        styles["background" if invert else "foreground"] = new
        return True

    return False


def parse_color(color: str, background: bool) -> str:
    """Parses a color tag."""

    background *= 10

    color = color.lstrip("@")

    if color.isdigit():
        index = int(color)

        if 0 <= index < 8:
            return str(30 + background + index)

        if 8 <= index < 16:
            index += 2  # Skip codes 38 & 48
            return str(80 + background + index)

        if index < 256:
            return f"{38 + background};5;{index}"

        raise ZmlNameError(
            color, "Color tags should be in the range 0-255.", expected_type="color"
        )

    # Substitute named colors when possible
    color = NAMED_COLORS.get(color, color)

    if color.startswith("#"):
        color = color.lstrip("#")

        color = ";".join(
            str(int(part, base=16)) for part in [color[:2], color[2:4], color[4:]]
        )

    return f"{38 + background};2;{color}"


def _apply_tag(tag: str, styles: StyleMap) -> None:
    """Applies the given tag to the style map.

    Note that this mutates the given map!
    """

    if tag == "/":
        styles.update(**_get_style_map())
        return

    is_unsetter = tag.startswith("/")
    is_background = tag.startswith("@")

    tag = tag.lstrip("/").lstrip("@")

    if tag in styles or tag in ["fg", "bg"]:
        if tag in ["fg", "bg"]:
            styles["foreground" if tag == "fg" else "background"] = None

        else:
            styles[tag] = not is_unsetter

        return

    layer = "background" if is_background else "foreground"

    if RE_COLOR.match(tag):
        styles[layer] = Color.from_ansi(parse_color(tag, is_background))
        return

    if tag in NAMED_COLORS:
        styles[layer] = Color.from_hex(NAMED_COLORS[tag]).as_background(is_background)

        return

    if tag.startswith("~"):
        styles["hyperlink"] = "" if is_unsetter else tag[1:]
        return

    raise ZmlNameError(tag)


def _parse_macro(tag: str) -> tuple[MacroType, tuple[str, ...]]:
    """Parses the given macro into its name and arguments."""

    args_start = tag.find("(")

    if args_start == -1:
        return tag, []

    args = tag[args_start + 1 : -1].split(",")

    return tag[:args_start], args


def _apply_prefix(tag: str, prefix: str) -> str:
    """Applies the prefix to the given tag.

    Most importantly, this function can recognize `@tags`, and insert the prefix to the
    correct spot (between `@` and the first letter).
    """

    if tag.startswith("@"):
        return "@" + prefix + tag[1:]

    if tag.startswith("!"):
        return "!" + prefix + tag[1:]

    return prefix + tag


@lru_cache(512)
def combine_spans(spans: tuple[Span, ...]) -> str:
    """Combines the given span iterable into optimized ANSI text."""

    styles = _get_style_map()

    buff = ""
    hyperlink = ""
    span: Span | None = None

    for span in spans:
        if span is FULL_RESET:
            buff += "\x1b[0m"
            styles = _get_style_map()
            continue

        new = {}
        unset = []

        for key, value in span.attrs.items():
            if key in ["text", "reset_after"]:
                continue

            if key == "hyperlink" and value not in ["", styles[key]]:
                new[key] = value
                continue

            if styles[key] != value:
                new[key] = value

                if not value and key != "hyperlink":
                    unset.append(key)

        if len(unset) > 0:
            # TODO: Maybe this could insert into the span's sequences?
            buff += "\x1b["

            for key in unset:
                buff += UNSETTERS[key] + ";"

            buff = buff.rstrip(";") + "m"

        buff += str(Span(text=span.text, reset_after=False, **new))
        styles.update(**new)

    return buff


@lru_cache(512)
def zml_get_spans(text: str) -> tuple[Span, ...]:
    """Gets all spans from the given ZML text."""

    styles: StyleMap = _get_style_map()
    spans: list[Span] = []

    for mtch in RE_MARKUP.finditer(text):
        tags, plain = mtch.groups()

        # No tags or plain value, likely at the end of the string.
        if tags == plain == None:
            continue

        tags = tags or ""
        plain = plain or ""

        for tag in tags.split():
            if tag == "/":
                spans.append(FULL_RESET)

            _apply_tag(tag, styles)

        auto_fg = _apply_auto_foreground(styles)

        spans.append(Span(plain, **styles))

        if auto_fg:
            styles["foreground"] = None

    return (*spans,)


def zml_pre_process(text: str, prefix: str = "", ctx: MarkupContext = None) -> str:
    """Applies pre-processing to the given ZML text.

    Currently, this involves three steps:

    - Substitute all aliases with their real values
    - Evaluate macros on the plain text
    - Eliminate groups that don't contain a plain part
    """

    ctx = ctx or GLOBAL_CONTEXT

    aliased = ""
    for mtch in RE_MARKUP.finditer(text):
        tags, plain = mtch.groups()

        if tags == plain == None:
            continue

        if tags is not None:
            aliased += "["

            for tag in tags.split():
                prefixed = _apply_prefix(tag, prefix)

                if prefixed in ctx["aliases"]:
                    tag = ctx["aliases"][prefixed]

                aliased += tag + " "

            aliased = aliased.rstrip(" ") + "]"

        if plain is not None:
            aliased += plain

    text = aliased

    macros: dict[str, tuple[MacroType, tuple[str, ...]]] = {}
    get_macro = ctx["macros"].get

    output = ""
    for mtch in RE_MARKUP.finditer(text):
        tags, plain = mtch.groups()

        # No tags or plain value, likely at the end of the string.
        if tags == plain == None:
            continue

        tags = tags or ""
        plain = plain or ""

        tag_list = tags.split()

        for tag in tag_list.copy():
            if tag == "/":
                macros.clear()
                continue

            if not tag.startswith(("!", "/!")):
                continue

            if tag.startswith("/!"):
                name = tag[1:]

                if name not in macros:
                    raise ZmlSemanticsError(
                        name,
                        f"Cannot unset a macro when that isn't set.",
                        expected_type="macro",
                    )

                del macros[name]

            else:
                name, args = _parse_macro(tag)
                prefixed = _apply_prefix(name, prefix)
                macro = get_macro(prefixed, None)

                if macro is None:
                    raise ZmlNameError(prefixed, expected_type="macro")

                macros[name] = (macro, args)

            tag_list.remove(tag)

        if len(tag_list) > 0:
            output += f"[{' '.join(tag_list)}]"

        if plain is not None:
            for macro, args in macros.values():
                plain = macro(plain, *args)

            output += plain

    return output.replace("][", " ")


def zml_escape(text: str) -> str:
    """Escapes non-intended ZML tags."""

    return text.replace("[", r"\[").replace("]", r"\]")


def _escape(text: str, replacements: tuple[str, str]) -> str:
    """Escapes ZML-syntax characters with the given replacements."""

    return text.replace("\[", replacements[0]).replace("\]", replacements[1])


def _unescape(text: str, replacements: tuple[str, str]) -> str:
    """Escapes ZML-syntax characters with the given replacements."""

    return text.replace(replacements[0], "[").replace(replacements[1], "]")


def zml(
    markup: str,
    prefix: str = "",
    ctx: MarkupContext | None = None,
    escape_replacements: tuple[str, str] = ("{{{{", "}}}}"),
) -> str:
    """Parses ZML markup into optimized ANSI text.

    DOCUMENT THIS
    """

    markup = _escape(markup, escape_replacements)

    # TODO: This step should be cached/done smarter. It takes ages!
    markup = zml_pre_process(markup, prefix=prefix, ctx=ctx)

    return _unescape(combine_spans(zml_get_spans(markup)), escape_replacements)



from time import strftime
from typing import Literal
from functools import lru_cache

from slate.color import Color
from slate.color_info import NAMED_COLORS

from .markup import zml_macro, parse_color


@zml_macro
def upper(text: str) -> str:
    """Returns `text.upper()`."""

    return text.upper()


@zml_macro
def lower(text: str) -> str:
    """Returns `text.lower()`."""

    return text.lower()


@zml_macro
def title(text: str) -> str:
    """Returns `text.title()`."""

    return text.title()


@zml_macro
def time(template: str, time_format: str = "%s") -> str:
    """Formats the template's `time` key as `time.strftime(time_format)`."""

    return template.format(time=strftime(time_format))


@zml_macro
@lru_cache
def gradient(
    text: str, origin: str, method: Literal["shade", "rainbow"] = "shade"
) -> str:
    """Adds a gradient into the given plain text.

    Args:
        origin: The start of the gradient.
        method: The type of gradient to apply.
    """

    is_background = origin.startswith("@")
    color = Color.from_ansi(parse_color(origin.lstrip("@"), is_background))

    if method == "shade":
        steps = [
            color.darken(4),
            color.darken(2),
            color,
            color.lighten(2),
            color.lighten(4),
        ]

    elif method == "rainbow":
        steps = [
            color,
            color.hue_shift(1 / 5),
            color.hue_shift(2 / 5),
            color.hue_shift(3 / 5),
            color.hue_shift(4 / 5),
        ]

    else:
        raise ValueError(r"Unknown gradient method {method!r}.")

    output = ""
    current_block = 0
    blocksize = max(round(len(text) / 5), 1)

    for i, char in enumerate(text):
        if i % blocksize == 0 and current_block < len(steps):
            color = steps[current_block]
            output += f"[{'@' if color.is_background else ' '}{color.hex}]"

            current_block += 1

        output += char

    return output + ("[/bg]" if is_background else "[/fg]")



from __future__ import annotations

from argparse import ArgumentParser

from .markup import zml


def command_zml(argv: list[str] | None = None) -> None:
    """Parses arguments for the ZML parser command."""

    parser = ArgumentParser(description="The ZML CLI interface.")

    parser.add_argument("markup", help="The markup to be parsed.")
    parser.add_argument(
        "-e", "--escape", help="Prints the output escaped.", action="store_true"
    )

    args = parser.parse_args(argv)

    output = zml(args.markup)

    if args.escape:
        print(repr(output))
    else:
        print(output)
